import re
from typing import List, Union

from docx.document import Document
from docx.opc.constants import RELATIONSHIP_TYPE as RT


def docx_regex_search(
    doc: Document,
    regex: Union[str, re.Pattern],
) -> List[str]:
    """Iterates through all paragraphs, tables, and hyperlinks in a word doc,
    and extracts any strings that match the regex.

    TODO: this needs tests, but it works nicely as far as I recall.
    """
    matches: List[str] = []

    # search paragraphs
    for par in doc.paragraphs:
        mo = re.fullmatch(regex, par.text.strip())
        if mo:
            matches.append(mo.string)

    # search tables
    for table in doc.tables:
        for row in table.rows:
            for cell in row.cells:
                mo = re.fullmatch(regex, cell.text.strip())
                if mo:
                    matches.append(mo.string)

    # search for links
    for rel in doc.part.rels:
        Rel = doc.part.rels[rel]
        if Rel.reltype == RT.HYPERLINK:
            text = Rel._target
            mo = re.fullmatch(regex, text)
            if mo:
                matches.append(mo.string)

    return matches



"""Teacher Heper CLI"""

import argparse
import code

from .sis import Sis


sis = Sis.read_cache() if Sis.cache_exists() else None


def find_student(name):
    if not sis:
        raise ValueError("fatal: cache does not exist")
    print(sis.find_student(name, threshold=60))


def find_parent(name):
    if not sis:
        raise ValueError("fatal: cache does not exist")
    print(sis.find_parent(name))


def shell():
    code.interact(
        local={
            "sis": sis,
        }
    )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--student", "-s", help="Lookup a student and print the result")
    parser.add_argument("--parent", "-p", help="Lookup a parent and print the result")
    parser.add_argument(
        "--new",
        action="store_const",
        const=True,
        help=(
            "Regenerate the database by parsing student.csv and parent.csv "
            "in the $HELPER_DATA directory."
        ),
    )

    args = parser.parse_args()

    if args.new:
        Sis.new_school_year().write_cache()
    elif args.student:
        find_student(args.student)
    elif args.parent:
        find_parent(args.parent)
    else:
        shell()


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from dataclasses import dataclass
import logging
from typing import List, Union

from .classroom_wrapper import GoogleClassroomApiWrapper
from teacherhelper.sis import Sis, Student

sis = Sis.read_cache()
logger = logging.getLogger(__name__)


@dataclass
class GradeResult:
    student: Student
    grade: int


@dataclass
class GradingContext:
    """Context that is injected into ClassroomGrader.grade_one

    All of the dicts are documented by Google:

    | member      | documentation                                                                                                          |
    | ----------- | ---------------------------------------------------------------------------------------------------------------------- |
    | course      | https://googleapis.github.io/google-api-python-client/docs/dyn/classroom_v1.courses.html                               |
    | assignment  | https://googleapis.github.io/google-api-python-client/docs/dyn/classroom_v1.courses.courseWork.html                    |
    | submissions | https://googleapis.github.io/google-api-python-client/docs/dyn/classroom_v1.courses.courseWork.studentSubmissions.html |
    | students    | https://googleapis.github.io/google-api-python-client/docs/dyn/classroom_v1.courses.students.html                      |
    """

    student: Student

    google_student: dict
    course: dict
    assignment: dict
    submission: dict


class ClassroomGrader(ABC, GoogleClassroomApiWrapper):
    """Light wrapper around GoogleClassroomApiWrapper to facilitate grading
    assignments."""

    def grade_all(self) -> List[GradeResult]:
        retval: List[GradeResult] = []

        for course, assignment, submission in self.traverse_submissions():
            try:
                google_student = self.get_student_profile(submission)
                name = google_student["name"]["fullName"]
                sis_student = sis.find_student(name)
                if not sis_student:
                    raise Exception("student not found")

                grade = self.grade_one(
                    GradingContext(
                        student=sis_student,
                        google_student=google_student,
                        course=course,
                        assignment=assignment,
                        submission=submission,
                    )
                )

                retval.append(GradeResult(student=sis_student, grade=grade))

                self.log_grade(name, grade, assignment)
            except Exception:
                student_name = self.get_student_profile(submission)["name"]["fullName"]
                logger.exception("grading failed for student named %s", student_name)
                logger.debug("assignment: %s", assignment)
                logger.debug("submission: %s", submission)

        return retval

    @abstractmethod
    def grade_one(self, context: GradingContext) -> Union[int, bool]:
        """Grade a single assignment. The return value can be boolean or int,
        depending on whether the assignment is numerically graded, or graded
        for completion only.

        Any exception thrown by this method will be handled in *grade_many*,
        and we will gracefully move on to the next student, so feel free to use
        exceptions for control flow, since those will get logged more nicely
        than quietly returning."""

    def is_form_submitted(self, submission: dict) -> bool:
        """Grading method for google forms. This assumes that the form grades
        were already imported, and it returns a boolean indicating completion
        (was the form submitted or not)"""
        return bool(submission.get("draftGrade"))

    def is_slideshow_changed(self, submission) -> bool:
        """Grading method for google slides (completion). Note that this always
        only checks the second slide, which is where I typically put my do-nows"""
        # ensure that there is exactly one attachment on the submission
        if "attachments" not in submission["assignmentSubmission"]:
            return False
        if len(submission["assignmentSubmission"]["attachments"]) < 1:
            return False
        elif len(submission["assignmentSubmission"]["attachments"]) != 1:
            logger.debug(submission)
            raise ValueError("can not grade. too many attachments")

        teacher, student = self.get_slides(
            submission,
            submission["assignmentSubmission"]["attachments"][0],
        )
        try:
            return teacher["slides"] != student["slides"]
        except KeyError:
            logger.exception("cannot compare slides")
            logger.debug("teacher %s", teacher)
            logger.debug("student %s", student)

            # fail upwards; mark the student as having done thd assignment
            return True

    def log_grade(self, name, grade, assignment):
        """Output the result of a grading action to the logger."""
        if isinstance(grade, bool):
            logger.debug(
                "{name} {status} {assignment_name}".format(
                    name=name,
                    status="completed" if grade else "did not complete",
                    assignment_name=assignment["title"],
                )
            )
        else:
            logger.debug(
                "{name} recieved grade of {grade} on {assignment_name}".format(
                    name=name, grade=grade, assignment_name=assignment["title"]
                )
            )



import re
import logging
from typing import Any, Generator, List

from .client import get_service


logger = logging.getLogger(__name__)


class GoogleClassroomApiWrapper:
    """Wrapper for the google classroom API, which also optionally makes calls
    to the drive and slides API"""

    def __init__(
        self,
        *,
        match_assignments: List[str],
        match_classrooms: List[str],
    ):
        """`services` is a dict with keys `classroom`, `slides`, and `drive`,
        referring to the service objects from the google client APIs. Classroom
        is the only required services. If slides or drive are not passed in,
        ValueErrors will be thrown at runtime if methods that use them are
        called and they are not present.

        *match_classrooms* is a list of regex patterns for classrooms we will
        traverse. *match_assignments* is the same, but for assignments within
        classrooms. We will always traverse all student submissions.
        """
        # google client libraries are highly dynamic, so explicitly casting to
        # Any supresses type warnings
        self.classroom: Any = get_service("classroom", "v1")
        self.drive: Any = get_service("drive", "v3")
        self.slides: Any = get_service("slides", "v1")

        self._match_pats = match_classrooms
        self._match_assgt = match_assignments

        # mapping of assignment ids to teacher content, to avoid repetitive
        # google drive exports
        # TODO: implement me! this is currently unused
        self._teacher_content_cache = {}

    def get_classrooms(self) -> List[dict]:
        """Return the classrooms that match self._match_pats.

        Returns a list of courseId's that correspond to those classrooms.
        """
        courses = self.classroom.courses().list().execute()
        return self._filter_courses(courses)

    def _filter_courses(self, courses):
        ret = []
        for c in courses["courses"]:
            for p in self._match_pats:
                if re.search(p, c["name"]) is not None:
                    ret.append(c)
        return ret

    def get_assignments(self, course) -> List[dict]:
        assignments = (
            self.classroom.courses().courseWork().list(courseId=course["id"]).execute()
        )
        while token := assignments.get("nextPageToken") is not None:
            assignments += (
                self.classroom.courses()
                .courseWork()
                .list(courseId=course, nextPageToken=token)
                .execute()
            )
        return self._filter_assignments(assignments)

    def _filter_assignments(self, assignments):
        ret = []
        for c in assignments["courseWork"]:
            for p in self._match_assgt:
                if re.search(p, c["title"]) is not None:
                    ret.append(c)
        return ret

    def get_submissions(self, assignment) -> List[dict]:
        submissions = (
            self.classroom.courses()
            .courseWork()
            .studentSubmissions()
            .list(courseWorkId=assignment["id"], courseId=assignment["courseId"])
            .execute()
        )
        while token := submissions.get("nextPageToken") is not None:
            submissions += (
                self.classroom.courses()
                .courseWork()
                .studentSubmissions()
                .list(
                    courseWorkId=assignment["id"],
                    courseId=assignment["courseId"],
                    nextPageToken=token,
                )
                .execute()
            )
        return submissions["studentSubmissions"]

    def download_file(self, file_id: str, mime_type):
        return self.drive.files().export(fileId=file_id, mimeType=mime_type).execute()

    def traverse_submissions(self) -> Generator[tuple[dict, dict, dict], None, None]:
        """Yields current classroom, assignment, and submission resource
        while traversing all courseWorkSubmissions."""
        for classroom in self.get_classrooms():
            for assignment in self.get_assignments(classroom):
                for submission in self.get_submissions(assignment):
                    yield classroom, assignment, submission

    def get_student_profile(self, submission) -> dict:
        return self.classroom.userProfiles().get(userId=submission["userId"]).execute()

    def get_slides(self, submission, file) -> tuple[dict, dict]:
        """Returns a tuple of two slideshow representations: the teacher slide,
        then the student slide"""
        student_slide = (
            self.slides.presentations()
            .get(presentationId=file["driveFile"]["id"])
            .execute()
        )
        parent_id = self.find_parent_id(submission, file)
        teacher_slide = (
            self.slides.presentations().get(presentationId=parent_id).execute()
        )

        return teacher_slide, student_slide

    def find_parent_id(self, submission, file) -> str:
        """For a given file in a submission, find the parent submission from
        which it is derived. For example, navigate to the teacher template
        from a google doc that a student submitted."""

        original_assignment = (
            self.classroom.courses()
            .courseWork()
            .get(courseId=submission["courseId"], id=submission["courseWorkId"])
            .execute()
        )

        # we need to try to find the matching file from the assignment to view
        # the teacher template, or parent file, of the student submission

        n_att = len(original_assignment["materials"])

        if n_att == 0:
            raise ValueError("original assignment has no attachments")

        # if there is only one attachment on the original, we don't need to do
        # any finegaling
        if n_att == 1:
            return original_assignment["materials"][0]["driveFile"]["driveFile"]["id"]

        # google classroom uses the naming convention:
        #     <student name> - <original file name>
        # <original file name> should therefore be a substring of the student
        # file's file name. We can try to find the original by iterating over
        # all attachments.
        filename = file["driveFile"]["title"]

        for attachment in original_assignment["materials"]:
            if attachment["driveFile"]["title"] in filename:
                return attachment["driveFile"]["id"]

        raise ValueError(f"could not get content for assignment {submission}")

    def get_content(self, submission, file, mime_type="text/plain") -> tuple[str, str]:
        """For a given assignment, return the content of a submission for a
        given attachment. This returns a tuple of the teacher content, then
        the student content."""

        student_content = self.download_file(file["driveFile"]["id"], mime_type)
        parent_id = self.find_parent_id(submission, file)
        teacher_content = self.download_file(parent_id, mime_type)
        return teacher_content, student_content



import csv
from pathlib import Path

from teacherhelper._data_dir import get_data_dir
from ._entities import Homeroom, ParentGuardian, Student


class OnCourseMixin:
    """This doesn't specifically interface with the OnCourse API, it just
    initializes the helper cache with the spreadsheet reports that I output
    from oncourse, using those specific headers."""

    @classmethod
    def new_school_year(cls):
        """Take a spreadsheet of student data and parent data, build the
        necessary relationships.

        Student data csv should include columns:

        - first name
        - last name
        - grade level
        - homeroom teacher
        - email address 1
        - birth date

        Parent data csv should include:

        - guardian first name
        - guardian last name
        - student first name
        - student last name
        - primary contact
        - guardian email address 1
        - guardian mobile phone
        - guardian phone
        - guardian work phone
        - comments
        - allow contact
        - student resides with
        - relation to student
        """
        student_data = Path(get_data_dir(), "students.csv")
        guardian_data = Path(get_data_dir(), "parents.csv")
        STUDENTS = {}
        HOMEROOMS = {}

        # various adapters from grade level representations to int
        to_number = {"4th Grade": 4, "5th Grade": 5, "6th Grade": 6, "7th Grade": 7}

        with open(student_data, "r", encoding="utf-8-sig") as csvfile:
            rd = csv.DictReader(csvfile)
            for row in rd:
                # convert keys to lowercase to normalize case
                row = {k.lower(): v for k, v in row.items()}
                student = Student(
                    {
                        "first_name": row.get("first name"),
                        "last_name": row.get("last name"),
                        "grade_level": to_number[row.get("grade level", "")],
                        "homeroom": row.get("homeroom teacher"),
                        "email": row.get("email address 1"),
                        "birthday": row.get("birth date"),
                    }
                )
                STUDENTS[student.name] = student
                if student.homeroom not in HOMEROOMS:
                    HOMEROOMS[student.homeroom] = Homeroom(
                        student.homeroom,
                        student.grade_level,
                        [student],
                    )
                else:
                    HOMEROOMS[student.homeroom].students.append(student)

        # instantiation
        self = cls(HOMEROOMS, STUDENTS, {})  # type: ignore
        with open(guardian_data, "r", encoding="utf8") as csvfile:
            rd = csv.DictReader(csvfile)
            for row in rd:
                # normalize headers to lowercase
                row = {k.lower(): v for k, v in row.items()}
                context = {
                    "first_name": row.get("guardian first name"),
                    "last_name": row.get("guardian last name"),
                    "student": (
                        row.get("student first name", "")
                        + " "
                        + row.get("student last name", "")
                    ),
                    "primary_contact": row.get("primary contact"),
                    "email": row.get("guardian email address 1"),
                    "mobile_phone": row.get("guardian mobile phone"),
                    "home_phone": row.get("guardian phone"),
                    "work_phone": row.get("guardian work phone"),
                    "comments": row.get("comments"),
                    "allow_contact": row.get("allow contact"),
                    "student_resides_with": row.get("student resides with"),
                    "relationship_to_student": row.get("relation to student"),
                }
                # find student object match
                student = self.find_student(context["student"])  # type: ignore
                if not student:
                    continue
                if student.name != context["student"]:
                    raise ValueError(
                        f"Integrity error. {student.name} does not equal "
                        + context["student"]
                    )
                context["student"] = student

                for k, v in context.items():
                    # clean phone numbers
                    if "phone" in k:
                        if "_" in v:
                            continue
                        nums = [l for l in v if l.isnumeric()]
                        if not nums:
                            continue
                        if len(nums) < 10:
                            continue
                        if len(nums) > 11:
                            context["comments"] += f"\n{k} is {v}"
                            continue
                        try:
                            phone_number = int("".join(nums))
                            context[k] = phone_number
                        except TypeError:
                            continue
                    # convert boolean fields to boolean
                    if k in [
                        "primary_contact",
                        "allow_contact",
                        "student_resides_with",
                    ]:
                        if "Y" in v:
                            v = True
                        elif "N" in v:
                            v = False
                        else:
                            raise ValueError(
                                f"Supposedly boolean field {k} could not"
                                "be converted into a boolean value."
                            )
                        context[k] = v

                parent = ParentGuardian(context)
                student.guardians.append(parent)
                if context["primary_contact"]:
                    student.primary_contact = parent

        return self



"""The document writer can read the rubric template at ./rubric.docx and build
the output document of all student results, which will be returned to students."""

from copy import deepcopy
from dataclasses import dataclass
import logging
from pathlib import Path
from typing import Sequence, Union, cast, Dict

from docx import Document
from docx.document import Document as DocumentType
from docx.oxml.ns import nsdecls
from docx.oxml import parse_xml

from teacherhelper.sis import Student
from .._data_dir import get_data_dir


logger = logging.getLogger(__name__)


@dataclass
class Page:
    student: Student
    rubric_grades: Dict[str, int]
    extra_rubric_fields: Union[Dict[str, str], None] = None
    notes: Union[str, None] = None


class RubricWriter:
    def __init__(
        self,
        *,
        template_doc: Union[Path, str, DocumentType],
        grade_to_col_mapping: Dict[int, int],
        cell_highlight_color: str = "CCCCCC",
    ):
        """*template_doc* is a path to your template. If you pass a string,
        it will be presumed to be relative to $HELPER_DATA.

        grade_to_col_mapping is a mapping between the grade and the
        corresponding rubric column, which is used during rubric cell shading.
        The right side of this mapping should be a zero-based index of a
        column in the table. Consider this example table:

        |               | 0           | 10          | 15          | 20          | total           |
        | ------------- | ----------- | ----------- | ----------- | ----------- | --------------- |
        | homework      | ...criteria | ...criteria | ...criteria | ...criteria | <homework>      |
        | classwork     | ...criteria | ...criteria | ...criteria | ...criteria | <classwork>     |
        | participation | ...criteria | ...criteria | ...criteria | ...criteria | <participation> |

        The correct mapping for this rubric would be:
            grade_to_col_mapping = {0: 1, 10: 2, 15: 3, 20: 4}

        *cell_highlight_color* can be any hex color code that you want to use
        for shading cells. By default, it's just gray.
        """

        if isinstance(template_doc, Path):
            self.doc = cast(DocumentType, Document(template_doc))
        elif isinstance(template_doc, str):
            # template_doc path is relative to teacherhelper data dir when a
            # string is passed
            self.doc = Document(get_data_dir() / template_doc)
        else:
            self.doc = template_doc

        self.cell_highlight_color = cell_highlight_color
        self.grade_to_col_mapping = grade_to_col_mapping

    def add_page(self, page: Page):
        """Sets page to self.current_page, then dispatches to page writing
        methods in the following order:

        1. _setup_page
        2. before_rubric
        3. add_rubric
        4. after_rubric
        """
        self.current_page = page
        self._setup_page()
        self.before_rubric()
        self._add_rubric()
        self.after_rubric()

    def add_pages(self, pages: Sequence[Page]):
        for page in pages:
            self.add_page(page)
            logger.debug("wrote page for %s", page.student.name)

    def _setup_page(self):
        """This is where we add the page break, so probably don't override me."""
        self.doc.add_page_break()

    def before_rubric(self):
        """Insert content on the page before the rubric"""
        self.doc.add_paragraph(f"Your Name: {self.current_page.student.name}")

    def after_rubric(self):
        """Insert content on the page after the rubric"""
        if self.current_page.notes is not None:
            self.doc.add_paragraph(f"Notes: {self.current_page.notes}")

    def _add_rubric(self):
        """Merge data from `Page` into the current page's rubric."""
        p = self.doc.add_paragraph()
        p._p.addnext(self._template_table)

        for target, value in self.current_page.rubric_grades.items():
            self._insert_grade(target, value)

        if self.current_page.extra_rubric_fields is not None:
            for target, value in self.current_page.extra_rubric_fields.items():
                self._replace_table_text(target, value)

    def _insert_grade(self, template_target: str, value: int):
        """Looks for the template target, enters the value as a string, but
        also looks up the value from self.grade_to_col_mapping to find which
        cell to shade, and shade that cell."""
        _, i_row, _ = self._replace_table_text(
            template_target, str(value), search_tables=[self.doc.tables[-1]]
        )
        i_cell = self.grade_to_col_mapping[value]
        cell = self.doc.tables[-1].rows[i_row].cells[i_cell]
        self._highlight_cell(cell)

    def _replace_table_text(self, search, replace: str, search_tables=None):
        """Returns the index of the table, row, and cell that the text was
        found in. If `search_tables` is specified, the returned table index
        refers to the passed-in list, not tables in the whole document."""
        if search_tables is None:
            search_tables = self.doc.tables
        for i, table in enumerate(search_tables):
            for j, row in enumerate(table.rows):
                for k, paragraph in enumerate(row.cells):
                    if search in paragraph.text:
                        # cursed, but this seems to replace the paragraph's
                        # inner text, preventing the formatting from being
                        # changed, and maintaining the text position formatting
                        # as being in the middle of the cell
                        paragraph.paragraphs[0].text = replace
                        return i, j, k

        raise Exception(f"could not perform replacement for {search}")

    def _highlight_cell(self, cell):
        shading_elm_1 = parse_xml(
            r'<w:shd {nsdecl} w:fill="{color}"/>'.format(
                nsdecl=nsdecls("w"), color=self.cell_highlight_color
            )
        )
        cell._tc.get_or_add_tcPr().append(shading_elm_1)

    @property
    def _template_table(self):
        return deepcopy(self.doc.tables[0]._tbl)



from copy import copy

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from teacherhelper._data_dir import get_data_dir


class ClientWrapper:
    """Wrapper for Google's Python client library, which uses the teacherhelper
    data directory to load credentials and store a token."""

    BASE_DIR = get_data_dir() / "google_oauth"

    CREDENTIALS = BASE_DIR / "credentials.json"
    TOKEN = BASE_DIR / "token.json"

    DEFAULT_SCOPES = [
        "https://www.googleapis.com/auth/userinfo.email",
        "https://www.googleapis.com/auth/userinfo.profile",
        "https://www.googleapis.com/auth/presentations.readonly",
        "openid",
        "https://www.googleapis.com/auth/classroom.courses.readonly",
        "https://www.googleapis.com/auth/classroom.courseworkmaterials.readonly",
        "https://www.googleapis.com/auth/classroom.student-submissions.students.readonly",
        "https://www.googleapis.com/auth/classroom.rosters.readonly",
        "https://www.googleapis.com/auth/classroom.profile.photos",
        "https://www.googleapis.com/auth/drive.readonly",
    ]

    def __init__(self, scopes=None):
        if not self.CREDENTIALS.exists():
            raise ValueError("missing oauth credentials")

        if scopes is None:
            scopes = copy(self.DEFAULT_SCOPES)
        self.scopes = scopes

    def get_credentials(self) -> Credentials:

        credentials = None
        if self.TOKEN.exists():
            credentials = Credentials.from_authorized_user_file(self.TOKEN, self.scopes)
        # If there are no (valid) credentials available, let the user log in.
        if not credentials or not credentials.valid:
            if credentials and credentials.expired and credentials.refresh_token:
                credentials.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    self.CREDENTIALS, self.scopes
                )
                credentials = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(self.TOKEN, "w") as token:
                token.write(credentials.to_json())

        return credentials

    def get_service(self, service, version):
        credentials = self.get_credentials()
        return build(service, version, credentials=credentials)


_client = ClientWrapper()


def get_service(service: str, version: str):
    return _client.get_service(service, version)


def get_credentials():
    return _client.get_credentials()



"""Before being called SIS (Student Information System), the sis module was
just called `helper`, and it was basically the full scope of the library. This
module adapts from the new API to that old one."""

from typing import Union
import warnings

from .sis import Sis, Student


class Helper(Sis):
    def __init__(self, *a, **kw):
        super().__init__(*a, *kw)
        warnings.warn(
            "Helper interface is deprecated. Use teacherhelper.Sis instead.",
            DeprecationWarning,
        )

    def find_nearest_match(
        self, student_name: str, threshold=90, **_
    ) -> Union[Student, None]:
        return self.find_student(student_name, threshold)

    def find_parent(self, parent_name: str) -> Union[Student, None]:
        result = super().find_parent(parent_name)
        if result:
            return result.student

    @classmethod
    def read_cache(cls, *a, **kw):
        warnings.warn(
            "Helper interface is deprecated. Use teacherhelper.Sis instead.",
            DeprecationWarning,
        )
        return super().read_cache(*a, **kw)


def get_helper() -> Helper:
    return Helper.read_cache()



import os

from .helper import Helper  # raises DeprecationWarning on instantiation
from .email_ import Email

RECCOMENDED_ENV = [
    "EMAIL_USERNAME",
    "EMAIL_PASSWORD",
]

for var in RECCOMENDED_ENV:
    if not os.getenv(var):
        print(
            f"Environment variable {var} is missing, which is necessary for "
            "certain functions within this module."
        )



"""Interface to ~/.teacherhelper or whatever path the user choose to store
data for this module via the `HELPER_DATA` environment variable."""

import os
from pathlib import Path


def get_data_dir() -> Path:
    # create a Path object representing the DATA_DIR, either a user provided path
    # or ~/.teacherhelper
    user_specified_data_dir = os.getenv("HELPER_DATA")
    if not user_specified_data_dir:
        return Path(os.path.expanduser("~"), ".teacherhelper")
    else:
        data_dir = Path(user_specified_data_dir)
        if not data_dir.exists():
            raise FileNotFoundError(f"path {data_dir} does not exist")

        return data_dir



import csv
import random
from typing import cast
from shutil import rmtree
import shelve
import datetime
from tempfile import mkdtemp
from pathlib import Path
from unittest.mock import patch

import pytest

from teacherhelper._data_dir import get_data_dir
from .fixtures import students_csv, parents_csv
from .._sis import Sis
from .._entities import ParentGuardian


@pytest.fixture
def helper(monkeypatch, students_csv, parents_csv):
    """This ultimately calls into the OnCourseMixin to create a new populated
    helper object."""
    dir = Path(mkdtemp())
    monkeypatch.setenv("HELPER_DATA", str(dir))
    for filename, data in (
        ("students.csv", students_csv),
        ("parents.csv", parents_csv),
    ):
        with open(dir / filename, "w") as fp:
            writer = csv.writer(fp)
            writer.writerows(data)

    yield Sis.new_school_year()
    rmtree(dir)


@pytest.fixture(params=(list(range(5))))
def random_student(request, helper):
    """10 random helper students"""
    random.seed(request.param)
    return random.choice(list(helper.students.values()))


@pytest.fixture(params=list(range(5)))
def random_parent(request, helper):
    random.seed(request.param)
    all_guardians = set(i for g in helper.students.values() for i in g.guardians)
    return random.choice(list(all_guardians))


def check_helper_equality(a: Sis, b: Sis) -> bool:
    """Rough __eq__ check for helper objects"""
    try:
        assert isinstance(a, Sis)
        assert isinstance(b, Sis)

        assert list(a.students.keys()) == list(b.students.keys())
        assert list(a.homerooms.keys()) == list(b.homerooms.keys())
        for key in a.homerooms:
            x = a.homerooms[key]
            y = b.homerooms[key]
            assert x.students == y.students

        return True
    except AssertionError:
        return False


def test_write_cache(helper):
    helper.write_cache()
    cache = get_data_dir() / "cache"

    # some file matching this pattern is in the folder. Different platforms
    # will call it different things
    assert any("cache" in str(i) for i in get_data_dir().iterdir())

    # the slice is because we need to change `cache.db` to just `cache` at
    # the end of the path for shelve
    with shelve.open(str(cache), "r") as db:
        # `date` points to a datetime object
        assert isinstance(db["date"], datetime.datetime)
        date = cast(datetime.datetime, db["date"])

        # `date` is *almost* right now
        assert (datetime.datetime.now() - date).seconds < 1

        cached_helper = cast(Sis, db["data"])
        check_helper_equality(cached_helper, helper)


def test_read_cache(helper):
    helper.write_cache()
    cached_helper = Sis.read_cache()
    check_helper_equality(helper, cached_helper)


def test_find_student(helper, random_student):
    name = random_student.name
    changed = list(name)
    changed[3] = "b"
    changed = "".join(changed)

    with patch("teacherhelper.sis._sis.process.extractOne") as p:

        # 89 is below the default threshold of 90
        p.return_value = name, 89
        result = helper.find_student(changed)
        assert p.mock_calls[-1].args[0] == changed
        assert result is None

        # if we lower the threshold, we get a result
        p.return_value = name, 89
        result = helper.find_student(changed, threshold=88)
        assert p.mock_calls[-1].args[0] == changed
        assert result is random_student

        # if we raise the confidence of the search result, we get a result
        p.return_value = name, 95
        result = helper.find_student(changed)
        assert p.mock_calls[-1].args[0] == changed
        assert result is random_student

        for call in p.mock_calls:
            assert call.args[1] == helper.students.keys()


def test_find_parent(helper, random_parent):
    name = random_parent.name
    result = helper.find_parent(name)
    assert result
    assert isinstance(result, ParentGuardian)

    # fuzzy matches work too
    chars = list(name)
    chars[1] = "c"
    chars[4] = "e"
    name = "".join(chars)
    result = helper.find_parent(name)
    assert result
    assert isinstance(result, ParentGuardian)

    # TODO: test that primary contacts are preferentially matched



from pathlib import Path

import os
import pytest

from teacherhelper.sis import Student
from .._rubric_writer import RubricWriter, Page


@pytest.fixture
def writer():
    return RubricWriter(
        template_doc=Path(Path(__file__).parent, "assets", "sample_rubric.docx"),
        grade_to_col_mapping={0: 1, 10: 2, 15: 3, 20: 4},
    )


@pytest.fixture
def guido():
    return Student({"first_name": "Guido", "last_name": "VanMorrison"})


@pytest.fixture
def timmy():
    return Student({"first_name": "Tim", "last_name": "Peters"})


@pytest.fixture
def joey():
    return Student({"first_name": "Joey", "last_name": "Jones"})


def test_add_page(writer, joey):
    writer.add_page(
        Page(
            student=joey,
            rubric_grades={"<week 19>": 10, "<week 20>": 15, "<week 21>": 0},
        )
    )

    # a table was added
    assert len(writer.doc.tables) == 2
    assert f"Your Name: {joey.name}" in [p.text for p in writer.doc.paragraphs]
    assert writer.doc.tables[1].rows[1].cells[-1].paragraphs[0].text == "10"
    assert writer.doc.tables[1].rows[2].cells[-1].paragraphs[0].text == "15"
    assert writer.doc.tables[1].rows[3].cells[-1].paragraphs[0].text == "0"


def test_page_with_notes(writer, timmy):
    writer.add_page(
        Page(
            student=timmy,
            rubric_grades={"<week 19>": 10, "<week 20>": 15, "<week 21>": 0},
            notes="You can do better than that, tim",
        )
    )
    assert (
        writer.doc.paragraphs[-1].text
        == "Notes: You can do better than that, tim"
        in [p.text for p in writer.doc.paragraphs]
    )


def test_add_pages(writer, joey, guido, timmy):
    writer.add_pages(
        [
            Page(
                student=joey,
                rubric_grades={"<week 19>": 10, "<week 20>": 15, "<week 21>": 0},
            ),
            Page(
                student=timmy,
                rubric_grades={"<week 19>": 10, "<week 20>": 15, "<week 21>": 0},
            ),
            Page(
                student=guido,
                rubric_grades={"<week 19>": 10, "<week 20>": 15, "<week 21>": 0},
            ),
        ]
    )

    assert len(writer.doc.tables) == 4
    # spot check data
    assert writer.doc.tables[1].rows[1].cells[-1].text == "10"
